/* GCodeMapper - maps movements of a g-code program from one surface to another, eg. plane to cylinder.
   Copyright (C) 2013 Martin Lederhilger <martin.lederhilger@gmx.at>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "CylinderGCodeMapper.h"
#include "GCodeReader.h"
#include "GCodeWriter.h"
#include "PreConditionCheck.h"
#include "MathUtility.h"

#include <sstream>
#include <cmath>
#include <cassert>
#include <boost/format.hpp>
#include <Eigen/Geometry>

namespace GCodeMapper
{
  using namespace GCodeReader;
  using namespace GCodeWriter;
  using namespace GCodeCommon;
  
  CylinderGCodeMapper::CylinderGCodeMapper(const Eigen::Vector3f& cylinderOrigin,
                                                           double cylinderDiameter,
                                                           double rapidFeedRate,
                                          const GCodeToStringMap& code2StringMap)
  : mCylinderOrigin(cylinderOrigin),
    mCylinderDiameter(cylinderDiameter),
    mRapidFeedRate(rapidFeedRate),
    mRapidMotionCode(GCodeToString(code2StringMap, GC_RAPID_MOTION)),
    mStraightFeedCode(GCodeToString(code2StringMap, GC_STRAIGHT_FEED)),
    mArcFeedCodeCW(GCodeToString(code2StringMap, GC_ARC_FEED_CW)),
    mArcFeedCodeCCW(GCodeToString(code2StringMap, GC_ARC_FEED_CCW)),
    mAbsoluteDistances(GCodeToString(code2StringMap, GC_ABSOLUTE_DISTANCES)),
    mIncrementalDistances(GCodeToString(code2StringMap, GC_INCREMENTAL_DISTANCES)),
    mAbsoluteArcCenters(GCodeToString(code2StringMap, GC_ABSOLUTE_ARC_CENTERS)),
    mRelativeArcCenters(GCodeToString(code2StringMap, GC_RELATIVE_ARC_CENTERS)),
    mSelectXYPlane(GCodeToString(code2StringMap, GC_SELECT_XY_PLANE)),
    mSelectXZPlane(GCodeToString(code2StringMap, GC_SELECT_XZ_PLANE)),
    mSelectYZPlane(GCodeToString(code2StringMap, GC_SELECT_YZ_PLANE)),
    mSavedZ(0.)
  {
    pcc<std::invalid_argument>(cylinderDiameter > 0., "cylinderDiameter must be greater than zero");
    pcc<std::invalid_argument>(rapidFeedRate > 0., "rapidFeedRate must be greater than zero");
  }
    
  bool CylinderGCodeMapper::willTransformLine(const std::string& line, const GCodeCommon::GCodeStateTracker& state) const
  {
    return state.doesCurrentLineTriggerMotion() ||                      //transform
           state.planeSelectionMode.hasStateChangedOnCurrentLine() ||   //filter out
           state.distanceMode.hasStateChangedOnCurrentLine()       ||   //filter out
           state.arcCenterMode.hasStateChangedOnCurrentLine();          //filter out
  }
    
  std::string CylinderGCodeMapper::transformLine(const std::string& line, const GCodeCommon::GCodeStateTracker& oldState,
                                                 const GCodeCommon::GCodeStateTracker& newState) const
  {
    const double segmentLength = 1.;
    assert(segmentLength > 0.);
    
    //Note: straight motions with the same z-coordinate can be modeled with a helix,
    //      straight motions with the same y-coordinate can be modeled with a straight line
    //      all other motions will be segmented into tiny lines.
    std::string transformedLine = line;
    
    removeWords(transformedLine, mSelectXYPlane, mSelectXZPlane, mSelectYZPlane);
    removeWords(transformedLine, mAbsoluteDistances, mIncrementalDistances);
    removeWords(transformedLine, mAbsoluteArcCenters, mRelativeArcCenters);
    
    if(newState.doesCurrentLineTriggerMotion())
    {
      removeWords(transformedLine, mRapidMotionCode, mStraightFeedCode, mArcFeedCodeCW, mArcFeedCodeCCW);
      removeWords(transformedLine, "X", "Y", "Z", "A", "B", "C", "I", "J", "K", "P");
                  
      if(newState.motionMode.getState() == MM_RAPID_MOTION ||
         newState.motionMode.getState() == MM_STRAIGHT_FEED)
      {
        if(newState.motionMode.getState() == MM_RAPID_MOTION)
        {
          if(transformedLine.length())
            transformedLine += " ";
          transformedLine += generateWordWithParameter<double>("F", mRapidFeedRate);
        }
        
        if(oldState.y.getState() == newState.y.getState())
        {
          std::vector<Eigen::Vector3f> points(1, Eigen::Vector3f(newState.x.getState(), newState.y.getState(), newState.z.getState()));
          transformedLine = writeStraightFeedLineSegments(transformedLine, points);
        }
        else if(oldState.z.getState() == newState.z.getState())
        {
          Eigen::Vector4f oldCylPt = transformPoint(Eigen::Vector3f(oldState.x.getState(), oldState.y.getState(), oldState.z.getState()));
          Eigen::Vector4f newCylPt = transformPoint(Eigen::Vector3f(newState.x.getState(), newState.y.getState(), newState.z.getState()));
          const std::string& mArcFeedCode = newCylPt[3] < oldCylPt[3] ? mArcFeedCodeCW : mArcFeedCodeCCW;
          transformedLine = boost::str(boost::format("%1% %2% %3% %4% %5% %6% %7% %8% %9%")
              % mArcFeedCode
              % transformedLine
              % generateWordWithParameter<double>("X", newCylPt[0])
              % generateWordWithParameter<double>("Y", newCylPt[1])
              % generateWordWithParameter<double>("Z", newCylPt[2])
              % generateWordWithParameter<double>("A", newCylPt[3])
              % generateWordWithParameter<double>("J", mCylinderOrigin[1])
              % generateWordWithParameter<double>("K", mCylinderOrigin[2])
              % generateWordWithParameter<int>   ("P", static_cast<int>(fabs(newCylPt[3]) / 360.)+1));
        }
        else
        {
          Eigen::Vector3f startPt(oldState.x.getState(), oldState.y.getState(), oldState.z.getState());
          Eigen::Vector3f endPt(newState.x.getState(), newState.y.getState(), newState.z.getState());
          Eigen::Vector3f diff = endPt - startPt;
          int segments = std::max(static_cast<int>(diff.norm() / segmentLength), 1);
          
          std::vector<Eigen::Vector3f> points;
          for(int i = 1; i <= segments; i++)
            points.push_back(startPt + diff * i / segments);
          transformedLine = writeStraightFeedLineSegments(transformedLine, points);
        }
      }
      else if(newState.motionMode.getState() == MM_ARC_FEED_CW || newState.motionMode.getState() == MM_ARC_FEED_CCW)
      {
        Eigen::Vector3f startPt(oldState.x.getState(), oldState.y.getState(), oldState.z.getState());
        Eigen::Vector3f endPt(newState.x.getState(), newState.y.getState(), newState.z.getState());
        Eigen::Vector3f arcCenter, arcAxis;
        const bool relArcCenters = newState.arcCenterMode.getState() == ACM_RELATIVE_ARC_CENTERS;
        
        switch(newState.planeSelectionMode.getState())
        {
          case PSM_SELECT_XY_PLANE:
            arcAxis =   Eigen::Vector3f::UnitZ();
            arcCenter = Eigen::Vector3f(newState.i.getState(), newState.j.getState(), relArcCenters ? 0. : endPt[2]);
            break;
          case PSM_SELECT_XZ_PLANE:
            arcAxis =   Eigen::Vector3f::UnitY();
            arcCenter = Eigen::Vector3f(newState.i.getState(), relArcCenters ? 0. : endPt[1], newState.k.getState());
            break;
          case PSM_SELECT_YZ_PLANE:
            arcAxis =   Eigen::Vector3f::UnitX();
            arcCenter = Eigen::Vector3f(relArcCenters ? 0. : endPt[0], newState.j.getState(), newState.k.getState());
            break;
          default:
            assert(false);
        }
        
        if(relArcCenters)
          arcCenter += startPt;
        
        Eigen::Vector3f arcCenterToStartPt = startPt - arcCenter;
        Eigen::Vector3f arcCenterToEndPt   = endPt   - arcCenter;
        double angle = MathUtility::angleBetweenVectors(arcCenterToStartPt, arcCenterToEndPt, arcAxis);
        if(angle < 0.)
          angle -= 2. * M_PI * (newState.p.getState() - 1.);
        else
          angle += 2. * M_PI * (newState.p.getState() - 1.);
        
        double rotAngle = 0.;
        if((angle >= 0. && newState.motionMode.getState() == MM_ARC_FEED_CCW) ||
           (angle < 0. && newState.motionMode.getState() == MM_ARC_FEED_CW)) //small circle
          rotAngle = angle;
        else //big circle
        {
          if(angle >= 0.)
            rotAngle = -2.*M_PI + angle;
          else
            rotAngle = 2.*M_PI + angle;
        }
        int segments = std::max(static_cast<int>(fabs(rotAngle) * arcCenterToStartPt.norm() / segmentLength), 1);
        std::vector<Eigen::Vector3f> points;
        for(int i = 1; i <= segments; i++)
        {
          Eigen::Affine3f t(Eigen::AngleAxisf(rotAngle * i / segments, arcAxis));
          points.push_back(arcCenter + t*arcCenterToStartPt);
        }
        
        transformedLine = writeStraightFeedLineSegments(transformedLine, points);
      }
    }
    
    return transformedLine;
  }
  
  std::string CylinderGCodeMapper::onEnableMapping(GCodeCommon::GCodeStateTracker& state)
  {
    std::stringstream sstream;
    
    //switch to absolute arc centers and select yz-plane for arcs.   
    sstream <<  boost::str(boost::format("%1% %2% %3%") % mAbsoluteDistances % mAbsoluteArcCenters % mSelectYZPlane) << std::endl;
    
    //this is also needed in onDisableMapping
    mSavedZ = state.z.getState();
    
    //we need to forge some movements to get an initial condition for mapping:    
    //this is for making the machines current z position over the cylinder surface available to the state tracker.
    std::string moveToSafeHeight = boost::str(boost::format("%1% %2%")
                                                % mRapidMotionCode
                                                % generateWordWithParameter<double>("Z", mSavedZ - mCylinderOrigin[2] - mCylinderDiameter / 2.));
    state.applyLine(moveToSafeHeight);
    
    //this is for moving to machine to the z-coordinate where we think we are. (eg. tool change macros, ...
    //could have been in between - changing the z-ccodinate)
    moveToSafeHeight = boost::str(boost::format("%1% %2%")
                                                % mRapidMotionCode
                                                % generateWordWithParameter<double>("Z", mSavedZ));
    sstream << moveToSafeHeight << std::endl;
    
    //this is for making the current y position to zero in the state tracker.
    std::string moveToYZero = boost::str(boost::format("%1% %2%")
                                                % mRapidMotionCode
                                                % generateWordWithParameter<double>("Y", 0.));
    state.applyLine(moveToYZero);
    
    //this is for moving the machine over the cylinder axis.
    moveToYZero = boost::str(boost::format("%1% %2%")
                                                % mRapidMotionCode
                                                % generateWordWithParameter<double>("Y", mCylinderOrigin[1]));
    sstream << moveToYZero;
    //end of forging some movements.
    
    return sstream.str();
  }
    
  std::string CylinderGCodeMapper::onDisableMapping(GCodeCommon::GCodeStateTracker& state)
  {
    std::stringstream sstream;
    
    //we need to forge some movements to get back to the condition before mapping:
    GCodeCommon::GCodeStateTracker tState1 = state;
    std::string moveToSafeHeight = boost::str(boost::format("%1% %2%")
                                                % mRapidMotionCode
                                                % generateWordWithParameter<double>("Z", mSavedZ - mCylinderOrigin[2] - mCylinderDiameter / 2.));
    state.applyLine(moveToSafeHeight);
    sstream << transformLine(moveToSafeHeight, tState1, state) << std::endl;
    
    std::string moveToYZero = boost::str(boost::format("%1% %2%")
                                                % mRapidMotionCode
                                                % generateWordWithParameter<double>("Y", 0.));
    GCodeCommon::GCodeStateTracker tState2 = state;
    state.applyLine(moveToYZero);
    sstream << transformLine(moveToYZero, tState2, state) << std::endl;
    //end of forging some movements.    
    
    //restore original arc center mode and arc plane.
    std::string restoreDistances, restoreArcCenters, restoreSelectedPlane;
    
    if(state.distanceMode.getState() == DM_ABSOLUTE_DISTANCES)
      restoreDistances = mAbsoluteDistances;
    else
      restoreDistances = mIncrementalDistances;
                                         
    if(state.arcCenterMode.getState() == ACM_ABSOLUTE_ARC_CENTERS)
      restoreArcCenters = mAbsoluteArcCenters;
    else
      restoreArcCenters = mRelativeArcCenters;
    
    switch(state.planeSelectionMode.getState())
    {
      case PSM_SELECT_XY_PLANE:
        restoreSelectedPlane = mSelectXYPlane;
        break;
      case PSM_SELECT_XZ_PLANE:
        restoreSelectedPlane = mSelectXZPlane;
        break;
      case PSM_SELECT_YZ_PLANE:
        restoreSelectedPlane = mSelectYZPlane;
        break;
      default:
        assert(false);
    }
    
    sstream << boost::str(boost::format("%1% %2% %3%") % restoreDistances % restoreArcCenters % restoreSelectedPlane);
    
    return sstream.str();
  }
    
  Eigen::Vector4f CylinderGCodeMapper::transformPoint(const Eigen::Vector3f point) const
  {
    Eigen::Vector4f tPoint;
    double alpha = point[1] / (mCylinderDiameter / 2.);
    double length = point[2] + mCylinderDiameter / 2.;
    pcc<std::invalid_argument>(length >= 0, "z coordinate of point is smaller than -radius");
        
    tPoint[0] = mCylinderOrigin[0] + point[0];
    tPoint[1] = mCylinderOrigin[1] + sin(alpha) * length;
    tPoint[2] = mCylinderOrigin[2] + cos(alpha) * length;
    tPoint[3] = -alpha / M_PI * 180.;
    
    return tPoint;
  }
  
  std::string CylinderGCodeMapper::writeStraightFeedLineSegments(std::string optionalCodes, const std::vector<Eigen::Vector3f>& points) const
  {
    std::stringstream sstream;
    
    for(std::vector<Eigen::Vector3f>::const_iterator iter = points.begin(), end = points.end(); iter != end; iter++)
    {
      Eigen::Vector4f cylPt = transformPoint(*iter);
      
      std::string line = boost::str(boost::format("%1% %2% %3% %4% %5% %6%")
            % mStraightFeedCode
            % optionalCodes
            % generateWordWithParameter<double>("X", cylPt[0])
            % generateWordWithParameter<double>("Y", cylPt[1])
            % generateWordWithParameter<double>("Z", cylPt[2])
            % generateWordWithParameter<double>("A", cylPt[3]));
      
      if(iter != points.begin())
        sstream << std::endl;
      
      sstream << line;
      
      if(iter == points.begin())
        optionalCodes = "";
    }
    
    return sstream.str();
  }
  
  const char* cCylinderOriginX = "cylinder-origin-x";
  const char* cCylinderOriginY = "cylinder-origin-y";
  const char* cCylinderOriginZ = "cylinder-origin-z";
  const char* cCylinderDiameter = "cylinder-diameter";
  const char* cRapidFeedRate = "rapid-feed-rate";
  
  namespace po = boost::program_options;
  
  std::string CylinderGCodeMapperBuilder::getName() const
  {
    return "CylinderGCodeMapper";
  }
  
  void CylinderGCodeMapperBuilder::addOptionsDescription(boost::program_options::options_description& desc) const
  {
    GCodeMapperBuilder::addOptionsDescription(desc);
    
    po::options_description mapperDesc("Cylinder mapper options");
    mapperDesc.add_options()
      (cCylinderOriginX,  po::value<float>()->default_value(0.)->required(), "Specifies the x coordinate of the cylinder's origin.")
      (cCylinderOriginY,  po::value<float>()->default_value(0.)->required(), "Specifies the y coordinate of the cylinder's origin.")
      (cCylinderOriginZ,  po::value<float>()->default_value(0.)->required(), "Specifies the z coordinate of the cylinder's origin.")
      (cCylinderDiameter, po::value<float>()->required(), "Specifies the diameter of the cylinder. Must be greater than zero.")
      (cRapidFeedRate,    po::value<float>()->required(), "Specifies the rapid feed rate of the machine.");
    desc.add(mapperDesc);
  }
  
  boost::shared_ptr<GCodeMapper> CylinderGCodeMapperBuilder::build(const boost::program_options::variables_map& variablesMap) const
  {
    Eigen::Vector3f cylinderOrigin(variablesMap[cCylinderOriginX].as<float>(),
                                   variablesMap[cCylinderOriginY].as<float>(),
                                   variablesMap[cCylinderOriginZ].as<float>());
    return boost::shared_ptr<GCodeMapper>(new CylinderGCodeMapper(cylinderOrigin,
                                                                  variablesMap[cCylinderDiameter].as<float>(),
                                                                  variablesMap[cRapidFeedRate].as<float>(),
                                                                  getGCodeToStringMapOption(variablesMap)));
  }
}
