/* GCodeMapper - maps movements of a g-code program from one surface to another, eg. plane to cylinder.
   Copyright (C) 2013 Martin Lederhilger <martin.lederhilger@gmx.at>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef CYLINDERGCODEMAPPER_H
#define CYLINDERGCODEMAPPER_H

#include "GCodeMapper.h"

#include <vector>
#include <Eigen/Eigen>

namespace GCodeMapper
{
  ///
  /// \brief This is the base class which applies certain transformations to a g-code program.
  class CylinderGCodeMapper : public GCodeMapper
  {
  public:
    EIGEN_MAKE_ALIGNED_OPERATOR_NEW
    
    /// \brief Maps a plane through the origin with the normal in z-direction to a cylinder
    ///        which has an axis of (1, 0, 0)'. The image of the point (0, 0, 0)' is cylinderOrigin +
    ///        cylinderRadius * (0, 0, 1)'.
    /// \param cylinderDiameter Diameter of the cylinder on which the plane is mapped. Must be greater than zero.
    /// \param rapidFeedRate G0 is emulated with other movements - a positive feed rate for this must be given.
    /// \param code2StringMap A reference to the map containing g-codes and their corresponding
    ///                       string representation.
    CylinderGCodeMapper(const Eigen::Vector3f& cylinderOrigin,
                                        double cylinderDiameter,
                                        double rapidFeedRate,
          const GCodeCommon::GCodeToStringMap& code2StringMap);
    
    virtual bool willTransformLine(const std::string& line, const GCodeCommon::GCodeStateTracker& state) const;
    
    virtual std::string transformLine(const std::string& line, const GCodeCommon::GCodeStateTracker& oldState,
                                      const GCodeCommon::GCodeStateTracker& newState) const;

    virtual std::string onEnableMapping(GCodeCommon::GCodeStateTracker& state);
    
    virtual std::string onDisableMapping(GCodeCommon::GCodeStateTracker& state);
    
  private:
    /// \brief Transforms a point from the plane around the cylinder.
    /// \param point The z coordinate of the point must be greater than -radius of cylinder.
    Eigen::Vector4f transformPoint(const Eigen::Vector3f point) const;
    
    /// \brief Generates straight feed line segments
    /// \param optionalCodes These codes are written on the first line after like "G01 <optionalCodes> ..."
    std::string writeStraightFeedLineSegments(std::string optionalCodes, const std::vector<Eigen::Vector3f>& points) const;
    
    const Eigen::Vector3f                mCylinderOrigin;
    double                               mCylinderDiameter,
                                         mRapidFeedRate;
    const std::string                    mRapidMotionCode,
                                         mStraightFeedCode,
                                         mArcFeedCodeCW,
                                         mArcFeedCodeCCW,
                                         mAbsoluteDistances,
                                         mIncrementalDistances,
                                         mAbsoluteArcCenters,
                                         mRelativeArcCenters,
                                         mSelectXYPlane,
                                         mSelectXZPlane,
                                         mSelectYZPlane;
    double                               mSavedZ;
  };
  
  ///
  /// \brief This class is a builder for the CylinderGCodeMapper.
  class CylinderGCodeMapperBuilder : public GCodeMapperBuilder
  {
  public:
    virtual std::string getName() const;
    virtual void addOptionsDescription(boost::program_options::options_description& desc) const;
    virtual boost::shared_ptr<GCodeMapper> build(const boost::program_options::variables_map& variablesMap) const;
  };
}

#endif