/* GCodeMapper - maps movements of a g-code program from one surface to another, eg. plane to cylinder.
   Copyright (C) 2013 Martin Lederhilger <martin.lederhilger@gmx.at>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "GCodeCommon.h"
#include "GCodeReader.h"
#include "PreConditionCheck.h"

namespace GCodeCommon
{
  using namespace GCodeReader;
  
  const std::string& GCodeToString(const GCodeToStringMap& map, GCode gcode)
  {
    GCodeToStringMap::const_iterator iter = map.find(gcode);
    pcc<std::invalid_argument>(iter != map.end(), "map does not contain all enum values");
    return iter->second;
  }

  GCodeStateTracker::GCodeStateTracker(const GCodeToStringMap& code2StringMap)
  : x("X", 0.), y("Y", 0.), z("Z", 0.),
    a("A", 0.), b("B", 0.), c("C", 0.),
    i("I", 0.), j("J", 0.), k("K", 0.),
    p("P", 1.),
    motionMode("", MM_RAPID_MOTION),
    planeSelectionMode("", PSM_SELECT_XY_PLANE),
    distanceMode("", DM_ABSOLUTE_DISTANCES),
    arcCenterMode("", ACM_RELATIVE_ARC_CENTERS),
    mCode2StringMap(code2StringMap)
  {
  }
    
  void GCodeStateTracker::applyLine(const std::string& lineIn)
  {
    std::string line = lineIn;
    const std::string& rapidMotionCode = GCodeToString(mCode2StringMap, GC_RAPID_MOTION);
    const std::string& straightFeedCode = GCodeToString(mCode2StringMap, GC_STRAIGHT_FEED);
    const std::string& arcFeedCodeCW = GCodeToString(mCode2StringMap, GC_ARC_FEED_CW);
    const std::string& arcFeedCodeCCW = GCodeToString(mCode2StringMap, GC_ARC_FEED_CCW);
    const std::string& absoluteDistancesCode = GCodeToString(mCode2StringMap, GC_ABSOLUTE_DISTANCES);
    const std::string& incrementalDistancesCode = GCodeToString(mCode2StringMap, GC_INCREMENTAL_DISTANCES);
    const std::string& absoluteArcCentersCode = GCodeToString(mCode2StringMap, GC_ABSOLUTE_ARC_CENTERS);
    const std::string& relativeArcCentersCode = GCodeToString(mCode2StringMap, GC_RELATIVE_ARC_CENTERS);
    const std::string& selectXYPlaneCode = GCodeToString(mCode2StringMap, GC_SELECT_XY_PLANE);
    const std::string& selectXZPlaneCode = GCodeToString(mCode2StringMap, GC_SELECT_XZ_PLANE);
    const std::string& selectYZPlaneCode = GCodeToString(mCode2StringMap, GC_SELECT_YZ_PLANE);
    
    motionMode.clearHasStateChangedOnCurrentLine();
    planeSelectionMode.clearHasStateChangedOnCurrentLine();
    distanceMode.clearHasStateChangedOnCurrentLine();
    arcCenterMode.clearHasStateChangedOnCurrentLine();
    
    //absolute or incremental distance mode
    if(hasWord<void>(absoluteDistancesCode, line))
    {
      distanceMode.setState(DM_ABSOLUTE_DISTANCES);
      extractWordParameterAndRemove<void>(absoluteDistancesCode, line);
    }
    else if(hasWord<void>(incrementalDistancesCode, line))
    {
      distanceMode.setState(DM_INCREMENTAL_DISTANCES);
      extractWordParameterAndRemove<void>(incrementalDistancesCode, line);
    }
    
    //absolute or relative arc centers
    if(hasWord<void>(absoluteArcCentersCode, line))
    {
      arcCenterMode.setState(ACM_ABSOLUTE_ARC_CENTERS);
      extractWordParameterAndRemove<void>(absoluteArcCentersCode, line);
    }
    else if(hasWord<void>(relativeArcCentersCode, line))
    {
      arcCenterMode.setState(ACM_RELATIVE_ARC_CENTERS);
      extractWordParameterAndRemove<void>(relativeArcCentersCode, line);
    }
    
    //plane selection
    if(hasWord<void>(selectXYPlaneCode, line))
    {
      planeSelectionMode.setState(PSM_SELECT_XY_PLANE);
      extractWordParameterAndRemove<void>(selectXYPlaneCode, line);
    }
    else if(hasWord<void>(selectXZPlaneCode, line))
    {
      planeSelectionMode.setState(PSM_SELECT_XZ_PLANE);
      extractWordParameterAndRemove<void>(selectXZPlaneCode, line);
    }
    else if(hasWord<void>(selectYZPlaneCode, line))
    {
      planeSelectionMode.setState(PSM_SELECT_YZ_PLANE);
      extractWordParameterAndRemove<void>(selectYZPlaneCode, line);
    }
    
    //movements
    if(hasWord<void>(rapidMotionCode, line))
    {
      motionMode.setState(MM_RAPID_MOTION);
      extractWordParameterAndRemove<void>(rapidMotionCode, line);
    }
    else if(hasWord<void>(straightFeedCode, line))
    {
      motionMode.setState(MM_STRAIGHT_FEED);
      extractWordParameterAndRemove<void>(straightFeedCode, line);
    }
    else if(hasWord<void>(arcFeedCodeCW, line))
    {
      motionMode.setState(MM_ARC_FEED_CW);
      extractWordParameterAndRemove<void>(arcFeedCodeCW, line);
    }
    else if(hasWord<void>(arcFeedCodeCCW, line))
    {
      motionMode.setState(MM_ARC_FEED_CCW);
      extractWordParameterAndRemove<void>(arcFeedCodeCCW, line);
    }
    
    //updateParameters
    updateParameters(line);   
  }
  
  bool GCodeStateTracker::doesCurrentLineTriggerMotion() const
  {
    return x.hasStateChangedOnCurrentLine() ||
           y.hasStateChangedOnCurrentLine() ||
           z.hasStateChangedOnCurrentLine() ||
           a.hasStateChangedOnCurrentLine() ||
           b.hasStateChangedOnCurrentLine() ||
           c.hasStateChangedOnCurrentLine();
  }
  
  void GCodeStateTracker::updateParameters(std::string& line)
  {
    bool useRelativeDistances = distanceMode.getState() == DM_INCREMENTAL_DISTANCES;
    updateParameter(line, x, useRelativeDistances);
    updateParameter(line, y, useRelativeDistances);
    updateParameter(line, z, useRelativeDistances);
    updateParameter(line, a, useRelativeDistances);
    updateParameter(line, b, useRelativeDistances);
    updateParameter(line, c, useRelativeDistances);
    updateParameter(line, i, false);
    updateParameter(line, j, false);
    updateParameter(line, k, false);
    updateParameter(line, p, false);
  }
  
  void GCodeStateTracker::updateParameter(std::string& line, WordState<double>& state, bool add)
  {
    state.clearHasStateChangedOnCurrentLine();
    
    if(hasWord<double>(state.getLetter(), line))
    {
      double value = extractWordParameterAndRemove<double>(state.getLetter(), line);
      if(add)
        state.setState(value + state.getState());
      else
        state.setState(value);
    }
  }
  
  namespace
  {
    void removeWord(std::string& line, const std::string& letter)
    {
      if(letter.length())
      {
        if(hasWord<void>(letter, line))
          extractWordParameterAndRemove<void>(letter, line);
        else if(hasWord<int>(letter, line))
          extractWordParameterAndRemove<int>(letter, line);
        else if(hasWord<double>(letter, line))
          extractWordParameterAndRemove<double>(letter, line);
      }
    }
  }
  
  void removeWords(std::string& line, const std::string& letter1,
                                      const std::string& letter2,
                                      const std::string& letter3,
                                      const std::string& letter4,
                                      const std::string& letter5,
                                      const std::string& letter6,
                                      const std::string& letter7,
                                      const std::string& letter8,
                                      const std::string& letter9,
                                      const std::string& letter10)
  {
    removeWord(line, letter1);
    removeWord(line, letter2);
    removeWord(line, letter3);
    removeWord(line, letter4);
    removeWord(line, letter5);
    removeWord(line, letter6);
    removeWord(line, letter7);
    removeWord(line, letter8);
    removeWord(line, letter9);
    removeWord(line, letter10);
  }
}
