/* GCodeMapper - maps movements of a g-code program from one surface to another, eg. plane to cylinder.
   Copyright (C) 2013 Martin Lederhilger <martin.lederhilger@gmx.at>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef GCODECOMMON_H
#define GCODECOMMON_H

#include <map>
#include <string>

namespace GCodeCommon
{
  enum GCode
  {
    GC_RAPID_MOTION,                     //G0
    GC_STRAIGHT_FEED,                    //G1
    GC_ARC_FEED_CW,                      //G2  //note g-code clock wise is inverted from the mathematical clockwise
    GC_ARC_FEED_CCW,                     //G3
    GC_ABSOLUTE_DISTANCES,               //G90
    GC_INCREMENTAL_DISTANCES,            //G91
    GC_ABSOLUTE_ARC_CENTERS,             //G90.1
    GC_RELATIVE_ARC_CENTERS,             //G91.1
    GC_UNIT_INCES,                       //G20
    GC_UNIT_MILLIMETERS,                 //G21      
    GC_SELECT_XY_PLANE,                  //G17
    GC_SELECT_XZ_PLANE,                  //G18
    GC_SELECT_YZ_PLANE,                  //G19
  };
  
  enum MotionMode
  {
    MM_RAPID_MOTION,
    MM_STRAIGHT_FEED,
    MM_ARC_FEED_CW,
    MM_ARC_FEED_CCW
  };
  
  enum PlaneSelectionMode
  {
    PSM_SELECT_XY_PLANE,
    PSM_SELECT_XZ_PLANE,
    PSM_SELECT_YZ_PLANE
  };
  
  enum DistanceMode
  {
    DM_ABSOLUTE_DISTANCES,
    DM_INCREMENTAL_DISTANCES
  };
  
  enum ArcCenterMode
  {
    ACM_ABSOLUTE_ARC_CENTERS,
    ACM_RELATIVE_ARC_CENTERS
  };
  
  typedef std::map<GCode, std::string> GCodeToStringMap;
  
  /// \brief Returns the string representation of the enum.
  /// \param map The map must contain all enums.
  /// \param gcode The enum value of the g-code for which the string is returned.
  /// \return \copybrief GCodeToString
  const std::string& GCodeToString(const GCodeToStringMap& map, GCode gcode);
  
  /// \brief This class keeps track of various g-code parameters.
  ///        throughout the program flow.
  class GCodeStateTracker
  {
  public:
    template<class T>
    class WordState
    {
    public:
      WordState(const std::string& letter, T initialState)
      : mLetter(letter),
        mState(initialState),
        mHasStateChangedOnCurrentLine(false)
      {
      }
      
      std::string getLetter() const
      {
        return mLetter;
      }
      
      T getState() const
      {
        return mState;
      }
      
      bool hasStateChangedOnCurrentLine() const
      {
        return mHasStateChangedOnCurrentLine;
      }
      
    private:
      friend class GCodeStateTracker;
      
      void setState(T state)
      {
        mState = state;
        mHasStateChangedOnCurrentLine = true;
      }
      
      void clearHasStateChangedOnCurrentLine()
      {
        mHasStateChangedOnCurrentLine = false;
      }
      
      std::string mLetter;
      T           mState;
      bool        mHasStateChangedOnCurrentLine;
    };
    
    /// \brief Initializes all states to zero.
    /// \param code2StringMap A reference to the map containing g-codes and their corresponding
    ///                       string representation.
    GCodeStateTracker(const GCodeToStringMap& code2StringMap);
    
    /// \brief Applies another g-code program line.
    /// \param line All parameters which can be found will be updated from this line.
    void applyLine(const std::string& line);
    
    /// \brief Returns the absolute x-coordinate.
    WordState<double> x;
    
    /// \brief Returns the absolute y-coordinate.
    WordState<double> y;
    
    /// \brief Returns the absolute z-coordinate.
    WordState<double> z;
    
    /// \brief Returns the absolute a-coordinate.
    WordState<double> a;
    
    /// \brief Returns the absolute b-coordinate.
    WordState<double> b;
    
    /// \brief Returns the absolute c-coordinate.
    WordState<double> c;
    
    /// \brief Returns the i-parameter.
    WordState<double> i;
    
    /// \brief Returns the j-parameter.
    WordState<double> j;
    
    /// \brief Returns the k-parameter.
    WordState<double> k;
    
    /// \brief Returns the p-parameter.
    WordState<double> p;
    
    /// \brief Returns the motion mode.
    WordState<MotionMode> motionMode;
    
    /// \brief Returns the selected plane for arcs.
    WordState<PlaneSelectionMode> planeSelectionMode;
    
    /// \brief Returns the distance mode.
    WordState<DistanceMode> distanceMode;
    
    /// \brief Returns the arc center mode.
    WordState<ArcCenterMode> arcCenterMode;
    
    /// \brief Returns true if the current line triggers a movement.
    /// \return \copybrief doesCurrentLineTriggerMotion
    bool doesCurrentLineTriggerMotion() const;
           
  private:
    void updateParameters(std::string& line);
    void updateParameter(std::string& line, WordState<double>& state, bool add);
    
    const GCodeToStringMap mCode2StringMap;
  };
  
  /// \brief Tries to remove words from a line.
  /// \note  No error is raised if a given word is not on the line.
  void removeWords(std::string& line, const std::string& letter1,
                                      const std::string& letter2 = std::string(),
                                      const std::string& letter3 = std::string(),
                                      const std::string& letter4 = std::string(),
                                      const std::string& letter5 = std::string(),
                                      const std::string& letter6 = std::string(),
                                      const std::string& letter7 = std::string(),
                                      const std::string& letter8 = std::string(),
                                      const std::string& letter9 = std::string(),
                                      const std::string& letter10= std::string());
}

#endif