/* GCodeMapper - maps movements of a g-code program from one surface to another, eg. plane to cylinder.
   Copyright (C) 2013 Martin Lederhilger <martin.lederhilger@gmx.at>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "GCodeMapper.h"
#include "PreConditionCheck.h"

namespace GCodeMapper
{
  GCodeMapper::~GCodeMapper()
  {
  }

  void mapGCode(std::istream& is, std::ostream& os, GCodeMapper& mapper,
                const std::string& beginMappingLine, const std::string& endMappingLine,
                const GCodeCommon::GCodeToStringMap& code2StringMap)
  {
    pcc<std::invalid_argument>(beginMappingLine.length() != 0, "beginMappingLine is an empty string");
    pcc<std::invalid_argument>(endMappingLine.length() != 0, "endMappingLine is an empty string");
    
    bool mappingEnabled = false;
    GCodeCommon::GCodeStateTracker newState(code2StringMap);
    
    is.exceptions(std::ios_base::badbit);
    os.exceptions(std::ios_base::badbit);
    
    while(!is.eof())
    {
      std::string line;
      std::getline(is, line);
      
      //get rid of possible carrige returns
      size_t foundPos = line.rfind('\r');
      if(foundPos != std::string::npos)
        line.erase(foundPos);
      
      if(line == beginMappingLine)
      {
        pcc<std::runtime_error>(!mappingEnabled, "encountered a mapping enable line although mapping has already been enabled");
        mappingEnabled = true;
        os << mapper.onEnableMapping(newState) << std::endl;
      }
      else if(line == endMappingLine)
      {
        pcc<std::runtime_error>(mappingEnabled, "encountered a mapping disable line although mapping has not been enabled");
        mappingEnabled = false;
        os << mapper.onDisableMapping(newState) << std::endl;
      }
      else
      {
        GCodeCommon::GCodeStateTracker oldState = newState;
        newState.applyLine(line);
        
        if(mappingEnabled && mapper.willTransformLine(line, newState))
          os << mapper.transformLine(line, oldState, newState) << std::endl;
        else
          os << line << std::endl;
      }
    }
  }
  
  const char *cBeginMappingLine = "begin-mapping-line";
  const char *cEndMappingLine = "end-mapping-line";
  const char* cGCRapidMotion = "gcode-rapid-motion";
  const char* cGCStraightFeed = "gcode-straight-feed";
  const char* cGCArcFeedCW = "gcode-arc-feed-cw";
  const char* cGCArcFeedCCW = "gcode-arc-feed-ccw";
  const char* cGCAbsoluteDistances = "gcode-absolute-distances";
  const char* cGCIncrementalDistances = "gcode-incremental-distances";
  const char* cGCAbsoluteArcCenters = "gcode-absolute-arc-centers";
  const char* cGCRelativeArcCenters = "gcode-relative-arc-centers";
  const char* cGCUnitInches = "gcode-unit-inches";
  const char* cGCUnitMillimeters = "gcode-unit-millimeters";
  const char* cGCSelectXYPlane = "gcode-select-xy-plane";
  const char* cGCSelectXZPlane = "gcode-select-xz-plane";
  const char* cGCSelectYZPlane = "gcode-select-yz-plane";
  
  namespace po = boost::program_options;
  
  GCodeMapperBuilder::~GCodeMapperBuilder()
  {
  }
    
  void GCodeMapperBuilder::addOptionsDescription(boost::program_options::options_description& desc) const
  {
    po::options_description mappingControlDesc("Mapping control");
    mappingControlDesc.add_options()
      (cBeginMappingLine, po::value<std::string>()->default_value("(begin mapping)")->required(), "Source code line which indicates that mapping should begin after this point in the program.")
      (cEndMappingLine,   po::value<std::string>()->default_value("(end mapping)")->required(), "Source code line which indicates that mapping should end after this point in the program.");
    desc.add(mappingControlDesc);  
    
    po::options_description gcodeDesc("G-codes");
    gcodeDesc.add_options()
      (cGCRapidMotion,          po::value<std::string>()->default_value("G0")->required(), "Specifies the rapid motion g-code.")
      (cGCStraightFeed,         po::value<std::string>()->default_value("G1")->required(), "Specifies the straight feed g-code.")
      (cGCArcFeedCW,            po::value<std::string>()->default_value("G2")->required(), "Specifies the clock wise arc feed g-code.")
      (cGCArcFeedCCW,           po::value<std::string>()->default_value("G3")->required(), "Specifies the coutner clock wise arc feed g-code.")
      (cGCAbsoluteDistances,    po::value<std::string>()->default_value("G90")->required(), "Specifies the absolute distances g-code.")
      (cGCIncrementalDistances, po::value<std::string>()->default_value("G91")->required(), "Specifies the incremental distances g-code.")
      (cGCAbsoluteArcCenters,   po::value<std::string>()->default_value("G90.1")->required(), "Specifies the absoulte arc centers g-code.")
      (cGCRelativeArcCenters,   po::value<std::string>()->default_value("G91.1")->required(), "Specifies the relative arc centers g-code.")
      (cGCUnitInches,           po::value<std::string>()->default_value("G20")->required(), "Specifies the select inches as unit g-code.")
      (cGCUnitMillimeters,      po::value<std::string>()->default_value("G21")->required(), "Specifies the select milimeter as unit g-code.")
      (cGCSelectXYPlane,        po::value<std::string>()->default_value("G17")->required(), "Specifies the select xy-plane for arc movements g-code.")
      (cGCSelectXZPlane,        po::value<std::string>()->default_value("G18")->required(), "Specifies the select xz-plane for arc movements g-code.")
      (cGCSelectYZPlane,        po::value<std::string>()->default_value("G19")->required(), "Specifies the select yz-plane for arc movements g-code.");
    desc.add(gcodeDesc);
  }
    
  std::string GCodeMapperBuilder::getBeginMappingOption(const boost::program_options::variables_map& variablesMap) const
  {
    return variablesMap[cBeginMappingLine].as<std::string>();
  }
  
  std::string GCodeMapperBuilder::getEndMappingOption(const boost::program_options::variables_map& variablesMap) const
  {
    return variablesMap[cEndMappingLine].as<std::string>();
  }
  
  GCodeCommon::GCodeToStringMap GCodeMapperBuilder::getGCodeToStringMapOption(const boost::program_options::variables_map& variablesMap) const
  {
    GCodeCommon::GCodeToStringMap codeToStringMap;
    {
      using namespace GCodeCommon;
      codeToStringMap[GC_RAPID_MOTION] = variablesMap[cGCRapidMotion].as<std::string>();
      codeToStringMap[GC_STRAIGHT_FEED] = variablesMap[cGCStraightFeed].as<std::string>();
      codeToStringMap[GC_ARC_FEED_CW] = variablesMap[cGCArcFeedCW].as<std::string>();
      codeToStringMap[GC_ARC_FEED_CCW] = variablesMap[cGCArcFeedCCW].as<std::string>();
      codeToStringMap[GC_ABSOLUTE_DISTANCES] = variablesMap[cGCAbsoluteDistances].as<std::string>();
      codeToStringMap[GC_INCREMENTAL_DISTANCES] = variablesMap[cGCIncrementalDistances].as<std::string>();
      codeToStringMap[GC_ABSOLUTE_ARC_CENTERS] = variablesMap[cGCAbsoluteArcCenters].as<std::string>();
      codeToStringMap[GC_RELATIVE_ARC_CENTERS] = variablesMap[cGCRelativeArcCenters].as<std::string>();
      codeToStringMap[GC_UNIT_INCES] = variablesMap[cGCUnitInches].as<std::string>();
      codeToStringMap[GC_UNIT_MILLIMETERS] = variablesMap[cGCUnitMillimeters].as<std::string>();
      codeToStringMap[GC_SELECT_XY_PLANE] = variablesMap[cGCSelectXYPlane].as<std::string>();
      codeToStringMap[GC_SELECT_XZ_PLANE] = variablesMap[cGCSelectXZPlane].as<std::string>();
      codeToStringMap[GC_SELECT_YZ_PLANE] = variablesMap[cGCSelectYZPlane].as<std::string>();
    }
    return codeToStringMap;
  }
}
