/* GCodeMapper - maps movements of a g-code program from one surface to another, eg. plane to cylinder.
   Copyright (C) 2013 Martin Lederhilger <martin.lederhilger@gmx.at>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef GCODEMAPPER_H
#define GCODEMAPPER_H

#include "GCodeCommon.h"

#include <istream>
#include <ostream>
#include <boost/program_options.hpp>
#include <boost/shared_ptr.hpp>

namespace GCodeMapper
{
  ///
  /// \brief This is the base class which applies certain transformations to a g-code program.
  class GCodeMapper
  {
  public:
    virtual ~GCodeMapper();
    
    /// \brief With this method a mapper implementation indicates, if it wants to transform this line.
    /// \param line g-code program line.
    /// \param state A reference to the g-code program state tracker, where current parameters can be retrieved.
    /// \return Returns true, if transform line will work with this line.
    virtual bool willTransformLine(const std::string& line, const GCodeCommon::GCodeStateTracker& state) const = 0;
    
    /// \brief Transforms this g-code program line.
    /// \param line g-code program line.
    /// \param newState A reference to the g-code program state tracker, where current parameters can be retrieved.
    /// \param oldState A reference to the g-code program state tracker, where previous parameters can be retrieved.
    virtual std::string transformLine(const std::string& line, const GCodeCommon::GCodeStateTracker& oldState,
                                      const GCodeCommon::GCodeStateTracker& newState) const = 0;
                                      
    /// \brief This method is called when mapping is enabled - this is when the mapper reads a beginMappingLine.
    /// \details The mapper can save state, and output lines.
    /// \param state A reference to the g-code program state tracker, where parameters can be retrieved. You
    ///              can also change the state of the tracker by appling new (generated) movements for example.
    /// \return The mapper can return g-code lines which are written into the output program.
    virtual std::string onEnableMapping(GCodeCommon::GCodeStateTracker& state) = 0;
    
    /// \brief This method is called when mapping is disabled - this is when the mapper reads a endMappingLine.
    /// \details The mapper can restore state, and output lines.
    /// \param state A reference to the g-code program state tracker, where parameters can be retrieved. You
    ///              can also change the state of the tracker by appling new (generated) movements for example.
    /// \return The mapper can return g-code lines which are written into the output program.
    virtual std::string onDisableMapping(GCodeCommon::GCodeStateTracker& state) = 0;
  };
  
  /// \brief This functions maps a g-code program.
  /// \param is The program is read from the input stream
  /// \param os and the transformed version is written to the output stream.
  /// \param mapper A reference to the mapper object.
  /// \param beginMappingLine If this line is encounterd in the input stream, mapping is enabled - Must not
  ///                      be a null string.
  /// \param endMappingLine If this line is encounterd in the input stream, mapping is disabled - Must not
  ///                       be a null string.
  /// \param code2StringMap A reference to the map containing g-codes and their corresponding
  ///                       string representation.
  /// \note  The mapper is only consulted if the program is in between beginMappingLine and endMappingLine.
  void mapGCode(std::istream& is, std::ostream& os, GCodeMapper& mapper,
                const std::string& beginMappingLine, const std::string& endMappingLine,
                const GCodeCommon::GCodeToStringMap& code2StringMap);
  
  ///
  /// \brief This is the base class for specifing program options and constructing GCodeMappers
  class GCodeMapperBuilder
  {
  public:
    virtual ~GCodeMapperBuilder();
    
    /// \brief Returns the unique name of the mapper.
    /// \note  This name is used to select on the command line which builder und thus which mapper is used.
    /// \return \copybrief getName
    virtual std::string getName() const = 0;
    
    /// \brief   Implementations have to add program options after the call to the base class.
    /// \details The base class contains program options which are used by all builders.
    /// \returns Program options which are later expected to show up in the build method.
    virtual void addOptionsDescription(boost::program_options::options_description& desc) const;
    
    /// \brief Builds the GCodeMapper object and returns a non null shared_ptr to it.
    /// \param variablesMap Containes the filled out program options.
    /// \returns \copybrief build
    virtual boost::shared_ptr<GCodeMapper> build(const boost::program_options::variables_map& variablesMap) const = 0;
    
    /// \brief Returns the value of the begin mapping program option.
    /// \return \copybrief getBeginMappingOption
    std::string getBeginMappingOption(const boost::program_options::variables_map& variablesMap) const;
    
    /// \brief Returns the value of the end mapping program option.
    /// \return \copybrief getEndMappingOption
    std::string getEndMappingOption(const boost::program_options::variables_map& variablesMap) const;
    
    /// \brief Returns the value of the gcodeToString program option.
    /// \return \copybrief getGCodeToStringMapOption
    GCodeCommon::GCodeToStringMap getGCodeToStringMapOption(const boost::program_options::variables_map& variablesMap) const;
  };
}

#endif