/* GCodeMapper - maps movements of a g-code program from one surface to another, eg. plane to cylinder.
   Copyright (C) 2013 Martin Lederhilger <martin.lederhilger@gmx.at>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "GCodeReader.h"
#include "PreConditionCheck.h"

#include <cassert>
#include <sstream>
#include <locale>
#include <boost/regex.hpp>
#include <boost/format.hpp>

namespace GCodeReader
{
  namespace
  {
    template<class ParamT>
    boost::regex formRegex(const std::string& letter);
    
    template<>
    boost::regex formRegex<void>(const std::string& letter)
    {
      pcc<std::invalid_argument>(letter.length(), "letter must not have zero length");
      
      return boost::regex(boost::str(boost::format(".*?(\\Q%1%\\E)(?: .*?|$)") % letter));
    }
    
    template<>
    boost::regex formRegex<int>(const std::string& letter)
    {
      pcc<std::invalid_argument>(letter.length(), "letter must not have zero length");
      
      return boost::regex(boost::str(boost::format(".*?(\\Q%1%\\E((?:\\+|-)?\\d+))(?: .*?|$)") % letter));
    }
    
    template<>
    boost::regex formRegex<double>(const std::string& letter)
    {
      pcc<std::invalid_argument>(letter.length(), "letter must not have zero length");
      
      return boost::regex(boost::str(boost::format(".*?(\\Q%1%\\E((?:\\+|-)?(?:\\d+|\\d+[.]\\d*|\\d*[.]\\d+)(?:(?:e|E)(?:\\+|-)?\\d+)?))(?: .*?|$)") % letter));
    }
    
    template<class ParamT>
    ParamT extractWordParameterImpl(const std::string& letter, std::string& line, bool remove)
    {
      boost::smatch match;
      pcc<std::runtime_error>(boost::regex_match(line, match, formRegex<ParamT>(letter)), "word is not in line");
      assert(match.size() == 3);
      
      std::stringstream sstream(match[2]);
      sstream.imbue(std::locale::classic());
      ParamT parameter;
      sstream >> parameter;
      
      if(remove)
      {
        line.erase(match[1].first - match[0].first, match[1].second - match[1].first + 1);
        if(!line.empty() && line[line.length() - 1] == ' ')
          line.erase(line.length() - 1);
      }
      
      return parameter;
    }
  }
  
  template<class ParamT>
  bool hasWord(const std::string& letter, const std::string& line)
  {
    boost::smatch match;
    return boost::regex_match(line, match, formRegex<ParamT>(letter));
  }
  
  template<class ParamT>
  ParamT extractWordParameter(const std::string& letter, const std::string& line)
  {
    return extractWordParameterImpl<ParamT>(letter, const_cast<std::string&>(line), false);
  }
  
  template<>
  void extractWordParameter<void>(const std::string& letter, const std::string& line)
  {
    
  }
  
  template<class ParamT>
  ParamT extractWordParameterAndRemove(const std::string& letter, std::string& line)
  {
    return extractWordParameterImpl<ParamT>(letter, line, true);
  }
  
  template<>
  void extractWordParameterAndRemove<void>(const std::string& letter, std::string& line)
  {
    boost::smatch match;
    pcc<std::runtime_error>(boost::regex_match(line, match, formRegex<void>(letter)), "word is not in line");
    assert(match.size() == 2);
    
    line.erase(match[1].first - match[0].first, match[1].second - match[1].first + 1);
    if(!line.empty() && line[line.length() - 1] == ' ')
      line.erase(line.length() - 1);
  }
  
  template bool hasWord<void>(const std::string&, const std::string&);
  template bool hasWord<int>(const std::string&, const std::string&);
  template bool hasWord<double>(const std::string&, const std::string&);
  
  template void   extractWordParameter<void>(const std::string&, const std::string&);
  template int    extractWordParameter<int>(const std::string&, const std::string&);
  template double extractWordParameter<double>(const std::string&, const std::string&);
  
  template void   extractWordParameterAndRemove<void>(const std::string&, std::string&);
  template int    extractWordParameterAndRemove<int>(const std::string&, std::string&);
  template double extractWordParameterAndRemove<double>(const std::string&, std::string&);
}
