/* GCodeMapper - maps movements of a g-code program from one surface to another, eg. plane to cylinder.
   Copyright (C) 2013 Martin Lederhilger <martin.lederhilger@gmx.at>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef MATHUTILITY_H
#define MATHUTILITY_H

#include "PreConditionCheck.h"

#include <Eigen/Eigen>
#include <cmath>

namespace MathUtility
{
  /// \brief Returns the angle between two vectors.
  /// \param   vec1 vector one - must not be a zero vector
  /// \param   vec2 vector two - must not be a zero vector
  /// \copybrief angleBetweenVectors
  template<class T>
  double angleBetweenVectors(const T& vec1, const T& vec2)
  {
    pcc<std::invalid_argument>(!vec1.isZero(), "v1 is a zero vector");
    pcc<std::invalid_argument>(!vec1.isZero(), "v2 is a zero vector");
    
    T unitVec1 = vec1 / vec1.norm();
    T unitVec2 = vec2 / vec2.norm();
    return acos(unitVec1.dot(unitVec2));
  }
  
  /// \brief Returns the angle between two vectors.
  /// \details The returned value is the angle by which vec1 must be rotated around axis to be aligned with vec2.
  /// \note    Note that this returns a signed angle.
  /// \param   vec1 vector one - must not be a zero vector and must be orthogonal to axis
  /// \param   vec2 vector two - must not be a zero vector and must be orthogonal to axis
  /// \param   axis rotation axis - must not be a zero vector
  /// \copybrief angleBetweenVectors
  template<class T>
  double angleBetweenVectors(const T& vec1, const T& vec2, const T& axis)
  {
    pcc<std::invalid_argument>(!vec1.isZero(), "v1 is a zero vector");
    pcc<std::invalid_argument>(!vec1.isZero(), "v2 is a zero vector");
    pcc<std::invalid_argument>(!axis.isZero(), "axis is a zero vector");
    pcc<std::invalid_argument>(fabs(axis.dot(vec1)) < 1.e-6, "vec1 must be orthogonal to axis");
    pcc<std::invalid_argument>(fabs(axis.dot(vec2)) < 1.e-6, "vec2 must be orthogonal to axis");
    
    T unitVec1 = vec1 / vec1.norm();
    T unitVec2 = vec2 / vec2.norm();
    if((unitVec2 - unitVec1).isZero())
      return 0.;
    else
    {
      T normal = unitVec2.cross(axis);
      
      double angle = acos(unitVec1.dot(unitVec2));
      if(unitVec1.dot(normal) >= 0.)
        return angle;
      else
        return -angle;      
    }
  }
}

#endif