/* GCodeMapper - maps movements of a g-code program from one surface to another, eg. plane to cylinder.
   Copyright (C) 2013 Martin Lederhilger <martin.lederhilger@gmx.at>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "GCodeMapper.h"
#include "CylinderGCodeMapper.h"
#include "PreConditionCheck.h"

#include <string>
#include <iostream>
#include <fstream>
#include <sstream>
#include <Eigen/Eigen>
#include <boost/format.hpp>
#include <boost/program_options.hpp>
#include <boost/make_shared.hpp>
#include <boost/filesystem.hpp>
#include <vector>

namespace po = boost::program_options;

typedef boost::shared_ptr<GCodeMapper::GCodeMapperBuilder> Builder;
typedef std::vector<Builder> Builders;

const char* cHelp = "help";
const char* cMapper = "mapper";
const char* cInputFile = "input-file";
const char* cOutputFile = "output-file";
const char* cConfigFile = "gcodemapper.conf";
const char* cStdIO = "-";

void showHelpMessage(const Builders& builders, const po::options_description& desc)
{
  std::cout << desc << std::endl;
  
  std::cout << "Available mappers:";
  for(Builders::const_iterator iter = builders.begin(), end = builders.end(); iter != end; iter++)
    std::cout << " " << (*iter)->getName();
  std::cout << std::endl;
}

Builder getBuilder(const Builders& builders, const std::string& name)
{
  for(Builders::const_iterator iter = builders.begin(), end = builders.end(); iter != end; iter++)
    if((*iter)->getName() == name)
      return *iter;
  throw std::runtime_error("Mapper name not found");
}

int main(int argc, char **argv)
{
  try
  {
    Builder builder;
    std::vector<boost::shared_ptr<GCodeMapper::GCodeMapperBuilder> > builders;
    builders.push_back(boost::make_shared<GCodeMapper::CylinderGCodeMapperBuilder>());
    
    po::variables_map vm;
    po::options_description desc("GCodeMapper V1.1 - maps movements of a g-code program from one surface to another.\n"
                                 "Copyright (C) 2013 Martin Lederhilger <martin.lederhilger@gmx.at>\n"
                                 "\n"
                                 "This program is free software: you can redistribute it and/or modify\n"
                                 "it under the terms of the GNU General Public License as published by\n"
                                 "the Free Software Foundation, either version 3 of the License, or\n"
                                 "(at your option) any later version.\n"
                                 "\n"
                                 "This program is distributed in the hope that it will be useful,\n"
                                 "but WITHOUT ANY WARRANTY; without even the implied warranty of\n"
                                 "MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n"
                                 "GNU General Public License for more details.\n"
                                 "\n"
                                 "You should have received a copy of the GNU General Public License\n"
                                 "along with this program.  If not, see <http://www.gnu.org/licenses/>\n"
                                 "\n"
                                 "All options");
    desc.add_options()
      (cHelp,                                 "Shows this help")
      (cMapper,     po::value<std::string>(), "Selects the mapper");
    if(boost::filesystem::exists(cConfigFile))
      po::store(po::parse_config_file<char>(cConfigFile, desc, true), vm);
    po::store(po::command_line_parser(argc, argv).options(desc).allow_unregistered().run(), vm);
    po::notify(vm);
    
    if(vm.count(cMapper))
    {
      builder = getBuilder(builders, vm[cMapper].as<std::string>());
      builder->addOptionsDescription(desc);
      desc.add_options()
        (cInputFile,  po::value<std::string>()->required(), "Input file path. Use '-' to read from stdin.")
        (cOutputFile, po::value<std::string>()->required(), "Output file path. Use '-' to write to stdout.");
        
      if(vm.count(cHelp))
        showHelpMessage(builders, desc);
      else
      {
        vm.clear();
        po::positional_options_description posDesc;
        posDesc.add(cInputFile, 1);
        posDesc.add(cOutputFile, 1);
        if(boost::filesystem::exists(cConfigFile))
          po::store(po::parse_config_file<char>(cConfigFile, desc), vm);
        po::store(po::command_line_parser(argc, argv).options(desc).positional(posDesc).run(), vm);
        po::notify(vm);
        
        boost::shared_ptr<GCodeMapper::GCodeMapper> mapper = builder->build(vm);
        pcc<std::runtime_error>(mapper, "mapper is a nullptr");
        
        std::string inFile = vm[cInputFile].as<std::string>();
        std::string outFile = vm[cOutputFile].as<std::string>();
        
        std::fstream fin;
        if(inFile != cStdIO)
          fin.open(inFile.c_str(), std::fstream::in);
        
        std::fstream fout;
        if(outFile != cStdIO)
          fout.open(outFile.c_str(), std::fstream::out | std::fstream::trunc);
        
        
        GCodeMapper::mapGCode(inFile == cStdIO ? std::cin : fin,
                              outFile == cStdIO ? std::cout : fout,
                              *mapper,
                              builder->getBeginMappingOption(vm),
                              builder->getEndMappingOption(vm),
                              builder->getGCodeToStringMapOption(vm));
      }
    }
    else
      showHelpMessage(builders, desc);
  }
  catch(const std::exception& e)
  {
    std::cerr << e.what() << std::endl;
    return 1;
  }
  
  return 0;
}

